;*****************************************************
;                                                   ;
;      BIOS60  MOPPEL-60-K-BIOS  V 12.62            ;
;                                                   ;
;*****************************************************
;
;   Stand:   Mo,02.03.87;19:43:10h
;
;
; Ueberarbeitet und Anpassung auf 8085Simulator-ID
;				https://www.oshonsoft.com/
;				notepad++ Tab = 6
;
; 01.07.2023 	Quellen von HMS (werner.roemer@t-online.de)
; 17.07.2023 	IDE-Karte als LW [D] eingebaut - nirawana
; 28.07.2023	IDE Einsprung berichtigt, trknr auf word erweitert
;			[D] mit 8MB
;
		org   0ea00h	;BIOS Start

;
; * Buffer kann um 1k noch oben geschoben werden
; 
buffer 	equ	f800h		; * FC00h	Buffer
verbuf 	equ	f700h		; * FB00h
variab 	equ	f500h		; * F900h
syspar 	equ	f400h		; * F800h Parameter aus Urloader
;
bios   	equ	ea00h		; 
bdos   	equ	dc06h
ccp    	equ	d400h
;
stacku 	equ	verbuf
stackc 	equ	 100h
stackw 	equ	  80h
buff   	equ	  80h
cdisk  	equ	   4h		;Kaltstart-Drive/User
iobyte 	equ	   3h
;
;
rambnk 	equ	  24h		;Banking-Latch
;      	equ
cimo   	equ	  43h		;Monitor-Unterprogramme
como   	equ	  49h
cstsm  	equ	  52h
cstsmo 	equ	100ch
lomo   	equ	  4fh
v24imo 	equ	1024h
v24omo 	equ	101eh
vinimo 	equ	  64h		;Adr.!
pinimo 	equ	  58h
;
stack0 	equ	2950h		;Bank-0-Stack
ioflag 	equ	2fc8h		;LO Select
;
;
begin  	equ	0101h		;Systemspuren ab Trk.1, Sec.1
nkalt  	equ	  1fh		;Anzahl Sektoren CCP/BDOS/BIOS
nwarm  	equ	  16h		;Anzahl Sektoren CCP/BDOS
offset 	equ	  04h		;Start bei Track 4
nmax   	equ	  04h		;LW-Anzahl
rwdhg  	equ	   5h		;Wiederholungen bei R/W-Fehler
timout 	equ	   6h		;max.Positionierzeit (x0,5 s)
;
;
		jmp  cboot 		;Kaltstart
		jmp  wboot 		;Warmstart
		jmp  const 		;Console Status
		jmp  conin 		;Console In
		jmp  conout		;Console Out
		jmp  list  		;Lineprinter
		jmp  punch 		;Puncher Out
		jmp  reader		;Reader In
		jmp  home  		;Track 0 einstellen
		jmp  seldsk		;Select Disk
		jmp  settrk		;Set Track No.
		jmp  setsec		;Set Sector No.
		jmp  setdma		;Set Disk Buffer
		jmp  read  		;Read Logical Sector
		jmp  write 		;Write Logical Sector
		jmp  lsts  		;Lineprinter Status
		jmp  sectra		;Get Translate Table
		jmp  prtini		;Schnittstellen initialisieren
		jmp  v24ini
;
;
;*****************************************************
;
;     Disk Parameter Header
;
;*****************************************************
;
dpbase:	ds   0h    		;Variablen und Tabellen am Ende
;
dpe0:  	dw   0h    		;Translate Table (XLT)
		dw   0h    		;Highest Dir.Entry
		dw   0h    		;Current Logical Track
		dw   0h    		;First Sector of Current Track
		dw   dirbuf		;Directory Buffer
		dw   dpb0  		;Disk Parameter Block
		dw   csv0  		;Checksum Vector
		dw   alv0  		;Allocation Vector
;
dpe1:  	dw   0h
		dw   0h
		dw   0h
		dw   0h
		dw   dirbuf
		dw   dpb1
		dw   csv1
		dw   alv1
;
dpe2:  	dw   0h
		dw   0h
		dw   0h
		dw   0h
		dw   dirbuf
		dw   dpb2
		dw   csv2
		dw   alv2
;
dpe3:  	dw   0h
		dw   0h
		dw   0h
		dw   0h
		dw   dirbuf
		dw   dpb3
		dw   csv3
		dw   alv3
;
;
;*****************************************************
;
;      Peripherie-Ansprache
;
;*****************************************************
;
conout:	lxi  h,cotab
		jmp  consol
conin: 	lxi  h,citab
consol:	lda  iobyte    ;CONSOLE (I/O-Bits 0&1)
		ani  3h
		jmp  decode
cotab: 	dw   co        ;Video Out
		dw   crtout
		dw   list
		dw   uscn3o
citab: 	dw   ci        ;ASCII In
		dw   crtin
		dw   reader
		dw   uscn3i
;
const: 	lda  iobyte    ;Console Status
		ani  3h
		lxi  h,cistab
		jmp  decode
cistab:	dw   csts      ;ASCII Status
		dw   crtst
		dw   ristat
		dw   uscn3s
;
;***
;
reader:	lda  iobyte    ;READER IN (I/O-Bits 2&3)
		rrc
		rrc
		ani  3h
		lxi  h,ritab
		jmp  decode
ritab: 	dw   ci        ;ASCII Input
		dw   v24in
		dw   usri2
		dw   usri3
;
ristat:	lda  iobyte    ;Reader Status
		rrc
		rrc
		ani  3h
		lxi  h,ristab
		jmp  decode
ristab:	dw   csts      ;ASCII Status
		dw   v24st
		dw   usri2s
		dw   usri3s
;
;***
;
punch: 	lda  iobyte    ;PUNCHER OUT (I/O-Bits 4&5)
		rrc
		rrc
		rrc
		rrc
		ani  3h
		lxi  h,potab
		jmp  decode
potab: 	dw   co        ;Video Out
		dw   v24out
		dw   uspo2
		dw   uspo3
;
pustat:	lda  iobyte    ;Puncher Status
		rrc
		rrc
		rrc
		rrc
		ani  3h
		lxi  h,pustab
		jmp  decode
pustab:	dw   rdysta    ;ready
		dw   v24st
		dw   uspo2s
		dw   uspo3s
;
;***
;
list:  	lda  iobyte    ;LISTER OUT (I/O-Bits 6&7)
		rlc
		rlc
		ani  3h
		lxi  h,lotab
		jmp  decode
lotab: 	dw   co        ;Video Out
		dw   crtout
		dw   prt
		dw   centr
;
lsts:  	lda  iobyte    ;Lister Status
		rlc
		rlc
		ani  3h
		lxi  h,listab
		jmp  decode
listab:	dw   rdysta    ;ready
		dw   crtst
		dw   prtst
		dw   censta
;
;
decode:	jz   found     ;Verzweigung nach Tabelle
		inx  h
		inx  h
		dcr  a
		jmp  decode
found: 	mov  e,m
		inx  h
		mov  d,m
		xchg           ;H&L: Zieladresse
		pchl
;
;
;*****************************************************
;
;      Monitor-Unterprogramme (incl.User)
;
;*****************************************************
;
monio: 	shld monadr    ;universelles Ein/Ausgabe-UP
		lxi  h,0h
		dad  sp        ;H&L: Bank 1 Stack
		lxi  d,stack0  ;D&E: Bank 0 Stack
		xchg
		sphl           ;Stack Pointer auf Bank 0 Stack
		mvi  a,81h
		out  rambnk    ;umschalten auf Bank 0
		push d         ;Bank 1 Stack retten
;
		db   0cdh      ;CALL
monadr:	ds   2h        ;UP-Startadresse
;
		mov  c,a
		pop  h         ;Bank 1 Stack zur}ck
		mvi  a,1h
		out  rambnk    ;Umschalten auf Bank 1
		sphl           ;alter Bank 1 Stack
		pop  h         ;Restore Regs
		pop  d
		mov  a,c
		ret
;
;
ci:    	push d         ;ASCII Input
		push h
		lxi  h,cisub
		jmp  monio
cisub: 	call cimo
		push psw
cilop: 	call cstsmo    ;auf Loslassen warten
		jnz  cilop
		pop  psw
		ret
;
;
co:    	push d         ;Video Output
		push h
		lxi  h,como
		jmp  monio
;
;
csts:  	push d         ;ASCII Status
		push h
		lxi  h,cstsm
		jmp  monio
;
;
prt:   	push d         ;Drucker (seriell)
		push h
		lxi  h,prtsub
		jmp  monio
prtsub:	mvi  a,80h     ;d.h. LO7
prtret:	sta  ioflag
		call lomo
		xra  a
		sta  ioflag
		ret
;
;
centr: 	push d         ;Drucker (parallel)
		push h
		lxi  h,censub
		jmp  monio
censub:	mvi  a,40h     ;d.h. LO6
		jmp  prtret
;
;
v24in: 	push d         ;V.24 Input
		push h
		lxi  h,v24imo
		jmp  monio
;
;
v24out:	push d         ;V.24 Output
		push h
		lxi  h,v24omo
		jmp  monio
;
;
v24ini:	push d         ;V.24 initialisieren
		push h
		lhld vinimo    ;Adr!
		jmp  monio
;
;
prtini:	push d         ;SER initialisieren
		push h
		lxi  h,pinimo
		jmp  monio
;
;
crtin: 	push d         ;Terminal Input
		push h
		lhld usadr0
		jmp  monio
;
;
crtout:	push d         ;Terminal Output
		push h
		lhld usadr1
		jmp  monio
;
;
uscn3i:	push d         ;User Console Input
		push h
		lhld usadr2
		jmp  monio
;
;
uscn3o:	push d         ;User Console Output
		push h
		lhld usadr3
		jmp  monio
;
;
usri2: 	push d         ;User Reader Input (1)
		push h
		lhld usadr4
		jmp  monio
;
;
usri3: 	push d         ;User Reader Input (2)
		push h
		lhld usadr5
		jmp  monio
;
;
uspo2: 	push d         ;User Puncher Out (1)
		push h
		lhld usadr6
		jmp  monio
;
;
uspo3: 	push d         ;User Puncher Out (2)
		push h
		lhld usadr7
		jmp  monio
;
;
prtst: 	ds 0h
censta:	ds 0h
v24st: 	ds 0h
crtst: 	ds 0h
uscn3s:	ds 0h
usri2s:	ds 0h
usri3s:	ds 0h
uspo2s:	ds 0h
uspo3s:	ds 0h
;
rdysta:	xra  a         ;Dummy Status: Device Ready
		dcr  a         ;A: FFh
		ret            ;ZERO off
;
;
;*****************************************************
;
;      Disketten-Management
;
;*****************************************************
;
home:  	lxi  b,0h      ; Track 0 einstellen
settrk:	push b         ; Trackbuffer laden
		pop  h           
		shld iotrk       
		ret            
;                            
;***                         
;                            
setsec:	push b         ; Sectorbuffer laden
		pop  h           
		shld iosec       
		ret            
;                            
;***                         
;                            
setdma:	push b         ; DMA-Adresse laden
		pop  h           
		shld iodma
		ret
;
;***
;
seldsk:	lxi  h,0h      ; fuer Fehlerbehandlung
		mov  a,c
		cpi  nmax      ; LW im Bereich?
		jc   selcte    ; ja:JMP
		lda  sysusr
		sta  cdisk
		ret            ; H&L=0 --> Select Fehler
;
selcte:	sta  dbank     ; LW-Nr.
		call parcop
		add  a
		add  a
		add  a
		add  a         ; DPB-Laenge: 16d
		mov  l,a
		lxi  d,dpbase
		dad  d         ; H&L: DPEx-Anfangsadresse
		ret
;
;***
;
sectra:	push b        ; in H&L Sektor-Nr.Uebergeben
		pop  h
		mov  a,d
		ora  e
		rz             ; No X-late Table
		dad  d
		mov  e,m       ; Wert aus Translate Table
		mvi  d,0h
		xchg
		ret
;
;***
;
read:  	call calcmp    ;Block lesen
		cnz  rload
		call bufcal
		call copy
		xra  a         ;o.k.:ACC=00h
		ret
;
rload: 	lda  wrtflg
		ora  a         ;Buffer leer?
		jz   rlod      ;ja: JMP
		call puttrk    ;sonst abspeichern
		lda  dbank
		call parcop
rlod:  	call gettrk
		ret
;
;***
;
write: 	mov  a,c       ;Block schreiben
		sta  dirflg    ;BDOS-FLAG
		call calcmp
		cnz  wload
		call bufcal
		xchg
		call copy
		mvi  a,1h
		sta  wrtflg    ;Buffer voll
		lda  dirflg
		cpi  1h        ;DIR-Record?
		mvi  a,0h
		rnz            ;nein: RET mit ACC=00h (=o.k.)
		call putsub
		xra  a
		sta  wrtflg    ;Buffer leer
		ret            ;o.k.:ACC=00h
;
wload: 	lda  wrtflg
		ora  a         ;Buffer leer?
		jz   wlod      ;ja: JMP
		call puttrk    ;sonst abspeichern
		lda  dbank
		call parcop
wlod:  	call gettrk
		ret
;
putsub:	call puttrk    ;wg.Error Handling
		ret
;
;
;====================================================
;
; Hier wird der Buffer von/auf Diskette oder IDE
;				geladen bzw. gesichert
;
; Test ob Laufwerk [D] ausgewählt wurde
;			dann zu den IDE-Routinen
;
;====================================================
;
gettrk:	push	psw
		mov	a,c	;
		cpi	03h		; Laufwerk = D ?
		jnz	get_fl	; nein weiter mit Floppy
		jmp	get_ide	;  ==> zu IDE-Routinen

;
puttrk:	push	psw
		mov	a,c	;
		cpi	03h		; Laufwerk = D ?
		jnz	put_fl	; nein weiter mit Floppy
		jmp	put_ide	;  ==> zu IDE-Routinen

;
; Diskettenroutine hier werden 4 Sektoren a256Byte
;					geladen, gesichert
;
put_fl:	pop	psw
		lda  aktdrv 	;Disk schreiben
		call parcop	
		lxi  h,secwr	
		jmp  pgsub	
;	
get_fl:	pop	psw	
		call calc   	;Disk lesen
		call store	
		lxi  h,secrd	
;	
pgsub: 	shld pgadr  	;PUT/GET-Unterprogramm
		call codbuf	
pglp:  	push h	
		push d	
		push b	
		call xlat	
		call save	
;	
		db   0cdh   	;CALL 
pgadr: 	ds   2h     	;FDC-Routine <-- lesen/schreiben
;	
		call resto	
		pop  b	
		pop  d	
		pop  h	
		jc   bdoser 	;CY on: Error
		call nxtsec	
		jnz  pglp	
		ret			; ==> zurueck zum BDOS
;
;
secrd: 	push h
		lxi  h,secred
		jmp  secsub
;
secwr: 	push h
		lxi  h,secwrt
;
secsub:	shld secadr
		pop  h
		shld hlbuff
		lxi  h,0h
		dad  sp
		shld spbuff
		lxi  sp,stacku
		lhld hlbuff
		db   0cdh      ;CALL
secadr:	ds   2h
		shld hlbuff
		lhld spbuff
		sphl
		lhld hlbuff
		ret
;
;
secred:	mvi  c,rwdhg
scrd:  	push h
		push d
		push b
		call secrdx
		pop  b
		pop  d
		pop  h
		rnc            ;CY off: o.k.
		dcr  c
		jnz  scrd
		stc            ;CY on:  Error
		ret
;
;
secwrt:	mvi  c,rwdhg
scwr:  	push h
		push d
		push b
		call secwrx
		pop  b
		pop  d
		pop  h
		jnc  verify    ;kein WR-Fehler, aber pr}fen!
		dcr  c
		jnz  scwr
		stc            ;CY on: Error
		ret
;
;
verify:	push b
		push d
		push h
		lxi  h,verbuf
		call secred    ;hier keine Fehlerauswertung
		pop  h
		push h
		call compar
		pop  h
		pop  d
		pop  b
		jnz  scwr
		ret            ;CY off (=o.k.): RET
;
;
;*****************************************************
;
;      Bufferverwaltung (Hilfsroutinen)
;
;*****************************************************
;
bufcal:	lxi  h,buffer
		lda  iosec
		ani  7h
		rrc
		mov  d,a
		mvi  a,0h
		rar
		mov  e,a
		mov  a,d
		ani  7fh
		mov  d,a
		dad  d
		xchg
		lhld iodma
		xchg
		mvi  c,80h     	;128 Bytes Laenge
		ret	
;	
codbuf:	xra  a         	;LW-Ansprache vorbereiten
		sta  wrtflg    	;Buffer leer
		lxi  h,buffer	; (HL)
		lda  aktsec		; 
		mov  e,a		; (E)
		lda  aktdrv		; <-- LW Zuordnung aufbereiten
		rlc
		ani  6h		; <-- Begrenzung auf 4LW
		mov  b,a
		mvi  d,4h      	; Zaehler fuer 4 Sektoren x256 Bytes
		lda  step		; Stepprate hinzufuegen
		ora  b
		mov  b,a
		lda  side		; Anzahl Seite testen
		ora  a
		rz			; --> Einseitig
		lda  trksid		; Track max (je nach Format)
		mov  c,a		;
		lda  akttrk
		cmp  c
		rc			; --> Anzahl im Bereich
		mov  a,b
		ori  20h
		mov  b,a
		ret
;
xlat: 	lda  side
		ora  a
		lda  akttrk
		mov  d,a
		rz
		cmp  c
		rc
		cma
		add  c
		add  c
		mov  d,a
		ret
;
; log. - in physikalische Sektoren wandeln
calc:  	lda  iosec
		rrc			; /2
		ani  0ch
		inr  a		; +1
		mov  e,a		; (E)
		lda  dbank		; ausgewaehltes LW
		mov  c,a		; (C)
		lda  iotrk		; ausgewaehlter Track
		mov  d,a		; (D)
		ret
;
; Parameter in Register ablegen
;
calcmp:	call calc		;
comp:  	lda  aktdrv
		cmp  c
		rnz
		lda  akttrk
		cmp  d
		rnz
		lda  aktsec
		cmp  e
		ret
;
compar:	mvi  c,0h      ;256d Bytes vergleichen
		push h
		lxi  h,verbuf
		xchg
		pop  h
comlop:	ldax d
		cmp  m
		rnz
		inx  h
		inx  d
		dcr  c
		jnz  comlop
		ret
;
store: 	mov  a,c
		sta  aktdrv
		mov  a,d
		sta  akttrk
		mov  a,e
		sta  aktsec
		ret
;
parcop:	push psw
		push h
		add  a
		add  a
		mov  e,a
		mvi  d,0h
		lxi  h,drv0
		dad  d
		lxi  d,step
		mvi  c,3h
		call copy
		pop  h
		pop  psw
		ret
;
copy:  	mov  a,m
		stax d
		inx  h
		inx  d
		dcr  c
		jnz  copy
		ret
;
nxtsec:	push d
		lxi  d,100h
		dad  d         ;Adresse +256d
		pop  d
		inr  e         ;Sektor-Nr.+1
		dcr  d         ;Sektor-Zaehler -1
		ret
;
save:  	lda  66h
		sta  nscbuf
		ret
;
resto: 	lda  nscbuf
		sta  66h
		ret
;
bdoser:	pop  psw       ;wg. 2xCALL
		pop  psw
		xra  a
		inr  a
		ret            ;Err.:ACC=01h
;
;
;*****************************************************
;
;      Floppy-Routinen
;
;*****************************************************
;
comreg   equ	48h
trkreg   equ	49h
secreg   equ	4ah
datreg   equ	4bh
status   equ	comreg
ctrl     equ	40h
;
;
intrd: 	in   datreg
		stax d
		inx  d
		ret
;
;
intwr: 	ldax d
		out  datreg
		inx  d
		ret
;
;
inton: 	push b
		mvi  a,1eh     ;8085-RST5.5 enable
		mvi  b,02h     ;NSC800-RSTC enable
		jmp  intcte
intof: 	push b
		mvi  a,1fh     ;8085-RST5.5 disable
		mvi  b,00h     ;NSC800-RSTC disable
intcte:	sim
		nop            ;wg.NSC800-JR-Befehl
		mov  a,b
		out  0bbh
		pop  b
		ret
;
;
init:  	di             ;Sektorzugriff vorbereiten
		push b
		shld datadr
		mvi  a,0e9h    ;PCHL
		sta  24h       ;8085-TRAP
		sta  66h       ;NSC-TRAP
		lxi  h,2ch     ;RST.5/RSTC
		mvi  m,0f1h    ;POP PSW
		inx  h
		mvi  m,0dbh    ;IN
		inx  h
		mvi  m,status
		inx  h
		mvi  m,0c9h    ;RET
;
		mov  a,d
		sta  trknr
		mov  a,e
		sta  secnr
		mov  a,b
		sta  selbyt
		rrc
		ani  3h
		mov  e,a
		mvi  d,0h
		lxi  h,lwnr
		dad  d
		mov  c,m
		mov  a,b
		ani  21h
		ora  c
		sta  selout
		lxi  h,track0
		dad  d
		pop  b
		ret
;
lwnr:  	db   02h
		db   04h
		db   08h
		db   10h
;
;
secwrx:	call init      ;Sektor schreiben
		push h
		lxi  h,intwr
		shld intprg
		pop  h
		lda  trknr
		mov  d,a
		mvi  a,2ah
		cmp  d         ;Sec.kleiner 42d?
		mvi  a,0a2h    ;*./.A6h; WRSEC
		jnc  scwrx
		mov  b,a
		lda  selbyt
		ani  01h
		mov  a,b
		jnz  scwrx
		ani  0fdh
scwrx: 	sta  rwbef
		jmp  start
;
;
secrdx:	call init      ;Sektor lesen
		push h
		lxi  h,intrd
		shld intprg
		pop  h
		mvi  a,80h     ;*./.84h; RDSEC (+30 ms)
		sta  rwbef
;
start: 	mov  a,m
		rlc
		cnc  restor    ;MSB off: Spur-Nr.ung}ltig
		mov  a,m
		ani  7fh
		out  trkreg
		lda  trknr
		out  datreg
		mvi  b,10h     ;*./.14h; SEEK
		call posit     ;Spur einstellen
		rc             ;CY on: RET (Fehler)
;
		mvi  c,rwdhg
wdhg:  	push b
		push h
		call rdwr      ;Schreib/Lese-Operation
		pop  h
		pop  b
		ani  5ch       ;Status auswerten
		jz   ende
		dcr  c
		jnz  wdhg
		jmp  fdcerr
;
ende:  	push psw
		call intof
		pop  psw
		ret
;
;
rdwr:  	lhld datadr
		push h
		pop  d
		lhld intprg
		lda  secnr
		out  secreg
		call inton
		lda  rwbef
		out  comreg
rdwrlp:	ei
		hlt
		jmp  rdwrlp
;
;
restor:	mvi  b,0h      ;*./.04h; RESTORE
;
posit: 	call pos       ;Spur einstellen
		ani  18h       ;Statusfehler?
		jnz  fdcerr    ;ja: JMP
		in   trkreg
		mov  c,a
		lda  trknr
		cmp  c         ;gew}nschte = aktuelle Spur?
		jnz  fdcerr    ;nein: JMP
fdcok: 	ori  80h       ;MSB on: Spur-Nr.g}ltig
		mov  m,a
		xra  a         ;CY off: kein Fehler
		ret
fdcerr:	mvi  m,0h      ;MSB off: Spur-Nr.ung}ltig
		stc            ;CY on: Fehler
		ret
;
pos:   	call inton     ;Kopf posit.;RET:Status im ACC
		call dely1     ;wg.Hitachi-Timing
		lda  selbyt
		rrc
		rrc
		rrc
		ani  3h        ;Step Rate
		ora  b         ;mit CMD verkn}pfen
		out  comreg
poslop:	in   status
		rlc            ;Motor on?
		jnc  poslop    ;nein: JMP
;
		lda  selout
		out  ctrl
		ei
;
		mvi  c,timout
timlop:	lxi  d,0h      ;TIMOUT x 0,52 s
timlp: 	dcx  d
		mov  a,d
		ora  e
		jnz  timlp
		dcr  c
		jnz  timlop
;
		mvi  a,0d0h    ;*(FORCE INTERRUPT)
		out  comreg    ;hier: Befehl abbrechen
		call intof
		mvi  a,0ffh    ;Pseudo-Status: Error
		ret
;
;
dely1: 	push psw
		mvi  a,0c0h    ;192d
delop: 	dcr  a
		jnz  delop
		pop psw
		ret
;
;
;****************************************************
;
;      Kalt- und Warmstart
;
;****************************************************
;
cboot: 	mvi  a,1h      ;Kaltstart
		out  rambnk    ;Bank 1 allein
		lxi  sp,stackc
;
		lxi  d,iobyte
		lxi  h,sysprt
		mov  a,m
		stax d         ;I/O-Byte laden
		inx  d
		inx  h
		mov  a,m
		stax d         ;Kalt-Disk/User laden
;
		mvi  a,0ffh
		sta  aktdrv    ;letztes Drive undefiniert
		mvi  a,0h
		sta  wrtflg    ;Buffer leer
		lxi  h,0h
		shld track0    ;Spur-Regs.undefiniert
		shld track2
;
		lxi  h,msg1    ;*"This is..."
		call msgout
;
gocpm: 	mvi  a,0c3h    ;Initialisierung Page 0
		sta  0h
		sta  5h
		lxi  h,bios
		lxi  d,3h
		dad  d         ;H&L: Warmstart-Adresse
		shld 1h
		lxi  h,bdos
		shld 6h
		lxi  b,buff
		call setdma
		lda  cdisk
		mov  c,a
		jmp  ccp
;
wboot: 	lxi  sp,stackw
		lda  wrtflg
		ora  a         ;ist alter Buffer gerettet?
		jz   load      ;ja: JMP
		call botsub    ;sonst abspeichern
		lda  dbank
		call parcop
;
load:  	mvi  a,0h
		sta  wrtflg    ;Buffer leer
		mvi  c,nwarm   ;nur (!) CCP und BDOS neu laden
		lxi  d,begin   ;Warmstart von Disk
		lxi  h,ccp
bloop: 	push d
		push h
		push b
		lda  drv0
		mov  b,a       ;Select-Byte
		call secrd
		pop  b
		pop  h
		lxi  d,0100h
		dad  d         ;Adresse +256d
		pop  d
		ora  a
		jc   load      ;Lesefehler: JMP
		inr  e
		dcr  c
		jz   gocpm     ;Start
;
		mov  a,e
		cpi  11h
		jnz  bloop
		inr  d         ;n{chste Spur,
		mvi  e,1h      ;erster Sektor
		jmp  bloop
;
;
botsub:	call puttrk    ;wg.Error Handling
		ret
;
;
;*****************************************************
;
;      Textausgabe
;
;*****************************************************
;
msgout:	mov  a,m
		ora  a
		rz
		mov  c,a
		push h
		call conout
		pop  h
		inx  h
		jmp  msgout
;
msg1:  	db   0dh
		db   0ah
		db   0ah
		db   "This is 60 K CP/M Rel. 2.2"
		db	0dh
		db	0ah
		db   "mit IDE-Interface V12.62"
		db   0dh
		db   0ah
		db   00h
;
;
;####################################################################
;
		org	0f100h	; Platz fuer IDE-Interface
;
;####################################################################
; Modul fuer IDE-Karte
;
; Entwickelt aus IDETOOL von Alfred Klichs 2022,;
; Paul's 8051 Code Library, Paul Stoffregen 2005
; https://www.pjrc.com/tech/8051/ide/
;
;
; Assembler 8085-Simulator-ID V4.93 Oshonsoft (Vladimir Soso)
; Editor 	Notepad++ TAB=6
;
;
; Stand:
;	24.01.2023	aus FDC-0123 vom 23.01.2023
;	28.07.2023 Beschreibung ergaenzt
;
;
; IDE-karte lesen/schreiben
;  Laufwerk 3 = IDE-Karte
;
; Datenübergabe:
;
;	akttrk <== iotrk BDOS SET-Track
;	aktsec <== iosec 1.Sektor in Moppelformat
;			 wird im IDE-Teil auf 512B/Sektor umgerechnet
;
;	wrtflg wird auf 0 gesetzt Buffer ungültig
;
; Parameteruebergabe:
;	buffer	= 1kByte fuer Daten
;	datbuf	= (word) aktuelle Bufferadresse fuer Zaehler
;	trknr		= (word)laufende Spurnummer
;	secnr		= (byte)laufende Sektornummer
;
;	LBA	= Sektor
;	LBA+1	= Track lsb
;	LBA+2	= Track msb
;	LBA+3	= Head
;
; Rückgabe:
;	Buffer 	= 1KByte Daten, werden dann Recordweise mit DMA ausgetauscht
;	(A)		= 0 für OK
;	(CY) 		= 1 Fehler lesen/schreiben
;
; msgout/string	= Textausgabe
;
;--------------------------------------------------------------------
;
; IO-Adressen fuer PIO 8255
;
PIO_IDE		EQU	058H		; PIO I/O Adresse
;
PIO_LSB		EQU	0 + PIO_IDE ; Port A
PIO_MSB		EQU	1 + PIO_IDE ; Port B
PIO_CTL		EQU	2 + PIO_IDE ; Port C
PIO_CFG		EQU	3 + PIO_IDE ; I/O modus
;
PIO_RDW		EQU	10010010B 	; Port A+B input, Port C output
PIO_WR		EQU	10000000B 	; Port A+B+C output
;
IDE_DRV		EQU	11100000B 	; select the device
			; x1x00000b LBA master
			; x1x10000b LBA slave
			; x0x00000b CHS master
			; x0x10000b CHS slave
;			; Kanal B als Ausgabe
;
;--------------------------------------------------------------------
;
; Variable fuer IDE Funktionen
;
IDE_A0		EQU	00000001B 	; A0 direct to IDE HDD
IDE_A1		EQU	00000010B 	; A1 dto.
IDE_A2		EQU	00000100B 	; A2 dto.
IDE_CS0		EQU	00001000B 	; CS0 to Inverter
IDE_CS1		EQU	00010000B 	; CS1 dto.
IDE_WR		EQU	00100000B 	; WR dto.
IDE_RD		EQU	01000000B 	; RD dto.
IDE_RST		EQU	10000000B 	; RST dto.
;
IDE_DATA		EQU	IDE_CS0
IDE_ERR		EQU	IDE_CS0 + IDE_A0
IDE_SEC_CNT		EQU	IDE_CS0 + IDE_A1
IDE_SECTOR		EQU	IDE_CS0 + IDE_A1 + IDE_A0
IDE_CYL_LSB		EQU	IDE_CS0 + IDE_A2
IDE_CYL_MSB		EQU	IDE_CS0 + IDE_A2 + IDE_A0
IDE_HEAD		EQU	IDE_CS0 + IDE_A2 + IDE_A1
IDE_COMMAND		EQU	IDE_CS0 + IDE_A2 + IDE_A1 + IDE_A0
IDE_STATUS		EQU	IDE_CS0 + IDE_A2 + IDE_A1 + IDE_A0
IDE_CONTROL		EQU	IDE_CS1 + IDE_A2 + IDE_A1
IDE_ASTATUS		EQU	IDE_CS1 + IDE_A2 + IDE_A1 + IDE_A0
;
IDE_CMD_RECAL	EQU	010H
IDE_CMD_READ	EQU	020H
IDE_CMD_WRITE	EQU	030H
IDE_CMD_INIT	EQU	091H
IDE_CMD_ID		EQU	0ECH
IDE_CMD_SPINDN	EQU	0E0H
IDE_CMD_SPINUP	EQU	0E1H
;
;
string		equ	msgout
;
;===========================================================
;
; IDE-karte lesen/schreiben
;  Laufwerk 3 = IDE-Karte
;
; Datenübergabe:
;
;	akttrk <== iotrk BDOS SET-Track
;	aktsec <== iosec 1.Sektor in Moppelformat
;			 wird im IDE-Teil auf 512B/Sektor umgerechnet
;
;	wrtflg wird auf 0 gesetzt Buffer ungültig
;
; Parameteruebergabe:
;	datbuf	= aktuelle Bufferadresse
;	trknr		= laufende Spurnummer
;	secnr		= laufende Sektornummer
;
;	LBA	= Sektor
;	LBA+1	= Track lsb
;	LBA+2	= Track msb
;	LBA+3	= Head
;
; Rückgabe:
;	Buffer 	= 1KByte Daten, werden dann Recordweise mit DMA ausgetauscht
;	(A)		= 0 für OK
;	(CY) 		= 1 Fehler lesen/schreiben
;
;--------------------------------------------------------------------
;
; initrw =	Buffer, Spur, Sektorregister vorbereiten
;
; CHS_LBA uebertraegt die CHS zu LBA
;
; Parameteruebergabe:
;
;	datbuf	= aktuelle Bufferadresse
;
;	LBA		= Sektor
;	LBA+1		= Track lsb
;	LBA+2		= Track msb
;	LBA+3		= Head

initrw:	xra  	a		; clear wrtflg -"Buffer ungültig"
		sta  	wrtflg	;
		di			; Interrupt sperren
		sta	lba+2
		sta	lba+3		; Head 0
		lxi	h,buffer	;
		shld	datbuf	;
		lhld	akttrk	;
		shld	trknr		;
		shld	lba+1		; Track lsb/msb
		lda	aktsec	;
		sta	secnr		;
		sta	lba		; Sektor
		lxi	h,buffer	;
		shld	datbuf	;
		ret			;
;
; get/put_cf =	Block von CF lesen/schreiben
;
get_ide:	pop	psw
		call 	initrw	;
		call 	READ_SECTOR	;	;
		ana	a		;
		jnz	cf_err	; Fehler
		;
		call 	updsek	; 2.Sektor laden
		call	READ_SECTOR	;
		ana	a		;
		jnz	cf_err	; Fehler
		xra	a		;
		ret			; zurück zum Disk-Manager
		;
		;
put_ide:	pop	psw
		call 	initrw	;
		call 	WRITE_SECTOR		;
		ana	a		;
		jnz	cf_err	; Fehler
		call 	updsek	;
		call	WRITE_SECTOR		;
		ana	a		;
		jnz	cf_err	; Fehler
		xra	a		;
		ret			; zurück zum Disk-Manager
;
;
; updsek/sp = Sektor und Spurnummern aktualisieren
;             und an CF-Karte senden
;
;	! maimal 32 Records pro Spur !
;
updsek:	lxi	h,buffer+512
		shld	datbuf
		lda	secnr		; Sektor holen
		cpi	16		; letzter Sektor spur
		jz	updsp		;
		inr	a		; nächster Sektor gleiche Spur
		sta	secnr		;
		sta	lba
		ret

updsp:	mvi	a,01h		; nächste Spur mit Sektor 1
		sta	secnr		;
		sta	lba
		lhld	trknr		;
		inx	h		;
		shld	trknr		;
		shld	lba+1		;
		ret
;
; Fehler (A)=1
;
cf_err:	mvi	a,01		;
		stc			;
		ret			; Zurück mit Fehler
;
;
;--------------------------------------------------------------------
;
ideinit:	call	IDE_HARD_RESET
		call	IDE_INIT
		jnz	nohd			; IDE Busy
		ret			;

nohd:		lxi	h,nomsg		; keine HD
		call	string		;
		ret

;
IDE_HARD_RESET:
		MVI	A, PIO_RDW		; PIO Read/Write Mode
		OUT	PIO_CFG
		MVI	A, IDE_RST		; hard reset drive
		OUT	PIO_CTL
		PUSH	B
		LXI	B,3000		; ca. 10 ms delay
WLOOP:
		DCX	B
		MOV	A,B
		ORA	C
		JNZ	WLOOP
		POP	B
		XRA	A			; no ide control lines asserted
		OUT	PIO_CTL
		RET
;
; HD Init
;
IDE_INIT:
		call	IDE_GET_ASTATUS	; Test ob Karte
		ana	a			; gesteckt ist
		jz	IDE_NOT_RDY		; nein --> Ende

		MVI	A,IDE_HEAD		;
		MVI	C,IDE_DRV		; select the device
		CALL	IDE_WR_8		;
		LXI	D,0010h		; wait for ready
IDE_RDY_BSY:
		MVI	B,10h
INILP:					; delay for slow drives
		DCR	B
		JNZ	INILP			; djnz $ Z80
		DCX	D
		MOV	A,D
		ORA	E
		JZ	IDE_NOT_RDY		; timeout
		MVI	A,IDE_STATUS
		CALL	IDE_RD_8
		MOV	A,C

		ANI	01000000B		; bit 6,a
						; Z80, if on then ready
		JZ	 IDE_RDY_BSY	; wait for READY bit to be set
		MOV	A,C

		ANI	10000000B		; bit 7,a
						; Z80, if on then busy
		JNZ	 IDE_RDY_BSY	; wait for BUSY bit to be clear
		RET

IDE_NOT_RDY:
		MVI	a,0FFh
		ORA	A
		STC				; timeout
		RET

;
;
; Wait for DRQ
;
IDE_WAIT_DRQ:
		MVI	A,0FFH
		MOV	B,A
		STA	DELAY_DRQ
MORE_DRQ:

		MVI	A,IDE_STATUS
		CALL	IDE_RD_8
		MOV	A,C
		ANI	10001000B
		CPI	00001000B
		JZ	DONE_DRQ
		DCR	B
		JNZ	MORE_DRQ
		LDA	DELAY_DRQ
		DCR	A
		JNZ	MORE_DRQ
		STA	DELAY_DRQ
		STC
		RET
DONE_DRQ:
		XRA	A
		RET

DELAY_DRQ:
		DB	0

;
; Wait for BUSY to be clear
;
IDE_WAIT_NOT_BUSY:
		MVI	A,0FFH
		MOV	B,A
		STA	DELAY_BUSY
MORE_BUSY:
		MVI	A,IDE_STATUS
		CALL	IDE_RD_8
		MOV	A,C
		ANI	11000000B
		XRI	01000000B
		JZ	NOT_BUSY
		DCR	B
		JNZ	MORE_BUSY
		LDA	DELAY_BUSY
		DCR	A
		STA	DELAY_BUSY
		JNZ	MORE_BUSY
		STC
		RET
NOT_BUSY:
		XRA	A
		RET

DELAY_BUSY:
		DB	0

;
; Get Active Status of the IDE Drive
;
IDE_GET_ASTATUS:
		MVI	A,IDE_ASTATUS
		CALL	IDE_RD_8
		MOV	A,C
		RET

;
; 8 bit read
;
IDE_RD_8:
		MOV	D,A			; save adress to D
		OUT	PIO_CTL		; drive address onto control lines
		ORI	IDE_RD		; set read pin on
		OUT	PIO_CTL
;		nop				; delay 2x nop
;		nop
		IN	PIO_LSB		; read lower byte
		MOV	C,A			; to C
		MOV	A,D			; reset read pin first
		OUT	PIO_CTL
		XRA	A			; reset all control pins
		OUT	PIO_CTL
		RET

;
; 8 bit write
;
IDE_WR_8:
		MOV	D,A			; save adress to D
						; Command, head, sector, cylinder etc. in C
		MVI	A, PIO_WR
		OUT	PIO_CFG		; PIO Write Mode
		MOV	A,C
		OUT	PIO_LSB		; Drive lower lines
		MOV	A,D			; Drive address onto control lines
		OUT	PIO_CTL
		ORI	IDE_WR		; set write pin on
		OUT	PIO_CTL
;		nop				; delay	2x nop
;		nop
		MOV	A,D
		OUT	PIO_CTL		; reset write pin first
		XRA	A
		OUT	PIO_CTL		; reset all control pins
		MVI	A,PIO_RDW
		OUT	PIO_CFG		; set PIO read/write mode
		RET

;
; Do the identify command and fill
; the buffer with the drive infos
;
DRIVE_ID:
		CALL	IDE_WAIT_NOT_BUSY
		MVI	A,IDE_HEAD
		MVI	C,IDE_DRV		; select the device
		CALL	IDE_WR_8
		CALL	IDE_WAIT_NOT_BUSY
		MVI	A,IDE_COMMAND
		MVI	C,IDE_CMD_ID	; 0ech
		CALL	IDE_WR_8		; issue the command
		CALL	IDE_WAIT_DRQ
		CALL	READ_DATA
		RET

;
; 16 bit read
;
READ_SECTOR:
						; hdread:
		CALL	IDE_WAIT_NOT_BUSY
		CALL	WR_LBA		; set sector
		MVI	A,IDE_COMMAND	; command register
		MVI	C,IDE_CMD_READ
		CALL	IDE_WR_8
		CALL	IDE_WAIT_DRQ
		CALL	READ_DATA
		CALL	IDE_WAIT_NOT_BUSY
		ANI	10001001B		; busy, DRQ or error
		RZ				; no
		JMP	GET_ERR

;
; 16 bit write
;
WRITE_SECTOR:
						; hdwrite:
		CALL	IDE_WAIT_NOT_BUSY
		CALL	WR_LBA		; set sector
		MVI	A,IDE_COMMAND	; command register
		MVI	C,IDE_CMD_WRITE
		CALL	IDE_WR_8
		CALL	IDE_WAIT_DRQ
		CALL	WRITE_DATA
		CALL	IDE_WAIT_NOT_BUSY
		ANI	10001001B		; busy, DRQ or error
		RZ				; no
						;
						; Return with errors if occurs
GET_ERR:
		MVI	A,IDE_ERR
		CALL	IDE_RD_8
		MOV	A,C
		ORA	A			; set Zero Flag
		RET

;
;
; Read one sector of 512 bytes
;




READ_DATA:	MVI	B,0			; 256 words, 512 bytes per sector
		lhld	datbuf		; target
READ_LP:
		MVI	A,IDE_DATA		; drive address onto control lines
		OUT	PIO_CTL		; and set read pin off (loop)
		ORI	IDE_RD
		OUT	PIO_CTL		; set read pin on
		IN	PIO_LSB		; read the lower byte
		MOV	M,A
		INX	H
		IN	PIO_MSB		; read the upper byte
		MOV	M,A
		INX	H
		DCR	B			; Z80 djnz read_lp
		JNZ	READ_LP
		XRA	A			; set all control pins of
		OUT	PIO_CTL		;
		shld	datbuf		; Zeiger sichern
		ret
;
; Write one sector of 512 bytes
;
WRITE_DATA:	MVI	B,0			; 256 words, 512 bytes per sector
		lhld	datbuf		; target			; source
		MVI	A, PIO_WR
		OUT	PIO_CFG		; PIO Write Mode
WRITE_LP:
		MOV	A,M
		OUT	PIO_LSB		; Drive lower lines with lsb
		INX	H
		MOV	A,M
		OUT	PIO_MSB		; Drive upper lines with msb
		MVI	A,IDE_DATA
		OUT	PIO_CTL		; Drive address onto control lines
		ORI	IDE_WR		; set write pin on
		OUT	PIO_CTL
		MVI	A,IDE_DATA
		OUT	PIO_CTL		; reset write pin
		INX	H
		DCR	B			; Z80 djnz write_lp
		JNZ	WRITE_LP
		XRA	A			; set all control pins of
		OUT	PIO_CTL
		MVI	A,PIO_RDW		; PIO Read/Write Mode
		OUT	PIO_CFG
		shld	datbuf		; Zeiger sichern
		RET
;
;
; LBA = 28bit 4+3x8
; If CHS mode head 8-bit (0-254), sektor 6-bit (1-63)
; cylinder/track 10-bit (0-1023) = 24-bit / 8,38GB max.
;
;
WR_LBA:	LDA	LBA+3			; 4/8-bit, CHS head 0-254
		ANI	0FH			; LBA 0-15
		ORI	IDE_DRV		; mode master/slave CHS/LBA
		MOV	C,A
		MVI	A,IDE_HEAD
		CALL	IDE_WR_8

		LDA	LBA+2			; 8-bit, CHS cyl msb
		MOV	C,A
		MVI	A,IDE_CYL_MSB
		CALL	IDE_WR_8

		LDA	LBA+1			; 8-bit, CHS cyl lsb
		MOV	C,A
		MVI	A,IDE_CYL_LSB
		CALL	IDE_WR_8

		LDA	LBA			; 8-bit, CHS sector

;		and	3fh			; 1-63 if CFC/CHS
		MOV	C,A
		MVI	A,IDE_SECTOR
		CALL	IDE_WR_8

		MVI	C,1			; transfer one sector
		MVI	A,IDE_SEC_CNT
		CALL	IDE_WR_8
		XRA	A
		RET
;
;
;
;--------------------------------------------------------------------
;
;
NOMSG:	DB	"No HDD found!"
		db	0ah		;
		db	0dh		;
		db	00h		;


;
;####################################################################
;
;
;*****************************************************
;
;      Variablen-Bereich
;
;*****************************************************
;
		org    syspar  ;vom Urloader
;
xlt0:  	ds   20h       ;Translate Tables
xlt1:  	ds   20h
xlt2:  	ds   20h
xlt3:  	ds   20h
;
dpb0:  	ds   10h       ;Disk Parameter Blocks
dpb1:  	ds   10h
dpb2:  	ds   10h
dpb3:  	ds   10h
;
sysprt:	ds   1h        ;I/O-Byte
sysusr:	ds   1h        ;Kaltstart-Drive/User
		ds  0eh
;
		ds   1h        ;Dive Parameter
drv0:  	ds   4h
drv1:  	ds   4h
drv2:  	ds   4h
drv3:  	ds   3h
;
sysjmp:	ds   1h        ;Codes umkopieren Y/N
sysvid:	ds   1h
		ds  0eh
;
usadr0:	ds   2h
usadr1:	ds   2h
usadr2:	ds   2h
usadr3:	ds   2h
usadr4:	ds   2h
usadr5:	ds   2h
usadr6:	ds   2h
usadr7:	ds   2h
;
;
		org    	variab
;
trknr: 	ds   2h        	; fuer FDC-Init
secnr: 	ds   1h		; 
;
selbyt:	ds   1h
selout:	ds   1h
datadr:	ds   2h
track0:	ds   1h 		;Spur-Regs.
track1:	ds   1h		
track2:	ds   1h		
track3:	ds   1h		
rwbef: 	ds   1h 		;RD/WR-Befehl
intprg:	ds   2h 		;RD/WR-Adresse
		ds   3h		
;		
dirflg:	ds   1h 		;von WRITE
wrtflg:	ds   1h 		;WR erl.:00h
nscbuf:	ds   1h		
hlbuff:	ds   2h		
spbuff:	ds   2h		
		ds   1h		
;		
dbank: 	ds   1h 		;SELDSK-Init.(LW-Nr.)
step:  	ds   1h		
side:  	ds   1h		
trksid:	ds   1h		
		ds   4h		
;		
iotrk: 	ds   2h 		;SETTRK
iosec: 	ds   2h 		;SETSEC
iodma: 	ds   2h 		;SETDMA
		ds   2h		
;		
aktdrv:	ds   1h 		;STORE-Err.
akttrk:	ds   1h
aktsec:	ds   1h
		ds   5h
;					; Fuer IDE Interface
lba:		ds	4		; Sektor
;					; Track lsb
;					; Track msb
;					; Head
;
datbuf:	dw   0h		;
;
;
dirbuf:	ds   80h       	;DIR Access Buffer
;	
alv0:  	ds   28h       	;27h fuer 2x80 Tracks
csv0:  	ds   20h       	;20h fuer 2x80 Tracks
alv1:  	ds   28h
csv1:  	ds   20h
alv2:  	ds   28h
csv2:  	ds   20h
alv3:  	ds   28h
csv3:  	ds   20h
;
;
end
;